/*
     This file is part of libextractor.
     (C) 2004 Vidyut Samanta and Christian Grothoff

     libextractor is free software; you can redistribute it and/or modify
     it under the terms of the GNU General Public License as published
     by the Free Software Foundation; either version 2, or (at your
     option) any later version.

     libextractor is distributed in the hope that it will be useful, but
     WITHOUT ANY WARRANTY; without even the implied warranty of
     MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
     General Public License for more details.

     You should have received a copy of the GNU General Public License
     along with libextractor; see the file COPYING.  If not, write to the
     Free Software Foundation, Inc., 59 Temple Place - Suite 330,
     Boston, MA 02111-1307, USA.

     This code was based on AVInfo 1.0 alpha 11 
     (c) George Shuklin, gs]AT[shounen.ru, 2002-2004
     http://shounen.ru/soft/avinfo/

 **/

#include "extractor_util.h"
#include "pack.h"
#include <string.h>
#include <stdlib.h>
#include <ctype.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <ctype.h>
#ifndef MINGW
#include <sys/mman.h>
#include <netinet/in.h>
#endif

#define RIFF_DEBUG 1
#define HAVE_MAIN 0 

static void addKeyword(struct EXTRACTOR_Keywords ** list,
		       char * keyword,
		       EXTRACTOR_KeywordType type) {
  EXTRACTOR_KeywordList * next;
  next = malloc(sizeof(EXTRACTOR_KeywordList));
  next->next = *list;    
  next->keyword = keyword;
  next->keywordType = type;
  *list = next;
}

#ifndef WINDOWS
  typedef unsigned int            DWORD;
  typedef unsigned short int      WORD;
  typedef signed int              LONG;
#endif

#define AVI_FPS_NUMBERS 100
/*100 for 2 digits, 1000 for 3, etc*/

#define AVI_MAX_STREAMS 9

enum{
  _FOURCC_RIFF=0x46464952,
  _FOURCC_AVI =0x20495641,
  _FOURCC_LIST=0x5453494C,
  _FOURCC_hdrl=0x6C726468,
  _FOURCC_AVIH=0x68697661,
  _FOURCC_strl=0x6C727473,
  _FOURCC_strh=0x68727473,
  _FOURCC_strf=0x66727473,
  _FOURCC_STRD=0x64727473,
  _FOURCC_vids=0x73646976,
  _FOURCC_auds=0x73647561,
  _FOURCC_INFO=0x4F464E49,
  _FOURCC_idx1=0x31786469,
  _FOURCC_movi=0x69766F6D,
  _FOURCC_JUNK=0x4B4E554A,
  _FOURCC_PAD =0x20444150,
  _FOURCC_DIV3=861292868,
  _FOURCC_DIVX=1482049860,
  _FOURCC_XVID=1145656920,
  _FOURCC_DX50=808802372,
  _FOURCC_fmt =0x20746D66,  /*for WAVE files*/
  _FOURCC_data=0x61746164,  /*for WAVE files*/
  _FOURCC_WAVE=0x45564157   /*for WAVE files*/
};

#define AVIIF_KEYFRAME  0x00000010L /*avi idx1 keyframe flag*/


typedef struct _var_desc{
  char* name;
  char* alias;
  int size;
} var_desc;


typedef struct _avioldindex_entry {
  unsigned char num1;  		/*|			*/
  unsigned char num2;  		/*+ unsigned char x4 = DWORD dwChunkId;*/
  unsigned char type;		/*|			*/
  unsigned char sub_type;	/*|			*/
  DWORD   dwFlags; 
  DWORD   dwOffset;
  DWORD   dwSize;
} aIndex2; /*some like aIndex, but more usefull*/

typedef struct _idx_stat{
  int frame_number;
  int size;
  int type;
} idx_stat;


/*offsets in strh array DWORD[]*/
const int fccType=2;
const int fccHanler=3;
const int dwScale=7;
const int dwRate=8;
const int dwLenght=10;

/*offsets in vids array DWORD[]*/
const int biWidth=3;
const int biHeight=4;
const int biBitCount=5;
const int biCompression=6;

/*offsets in auds array WORD[]   N.B. WORD != DWORD */
const int wFormatTag=4;
const int nChannels=5;
const int nSamplesPerSec=6;
const int wBitsPerSample=11;




static struct EXTRACTOR_Keywords * riffparse_INFO(char * buffer,
						  size_t size,
						  struct EXTRACTOR_Keywords * prev) {
  char backup;
  size_t c = 0;
  char type[5]="\0\0\0\0";
  int ok;
  char * pos;

  if (1)
    return prev; /* for now; format detection is too brittle at this point;
		    we get garbage results for various mp3-file formats that
		    are not properly detected otherwise! */
  if (size < 64)
    return prev;
  backup = buffer[size-4];
  buffer[size-4]='\0';

  do {
    memcpy(type,
	   buffer+c,
	   4); /* smartass copying of description name*/
    c += 8; /* now c points to description content */

    /*
      ???
      SetIdxStringVar("stream.name",desc_count,strdup(type));
    */
    
    /*
      printf("WHAT IS %s\n", type); ??
    */
    
    pos = buffer+c;
    ok = 1;
    while (pos[0] != '\0') {
      ok = ok && isprint(pos[0]);
      pos++;
    }
    if (ok) {    
      addKeyword(&prev,
		 strdup(buffer+c),
		 EXTRACTOR_UNKNOWN); /* eh, what exactly is it */
    }
    do {
      c++;
    } while (buffer[c] && c < size-8);
    c++;
    while (!buffer[c] && c<size-8)
      c++; 
  } while (c<size-8);

  buffer[size-4] = backup;
  return prev;
}

#ifdef later
static int riffparse_idx1 (FILE* file, int size){
  idx_stat *index;
  aIndex2 idx_entry;
  int tmp;
  int c, /*counter*/streams=0, str_num;/*current stream number*/
  int a_c=0, v_c=0;

  index = calloc( sizeof(idx_stat) , AVI_MAX_STREAMS );
  for(c=0;c<size/sizeof(aIndex2);c++){
    SAFEREAD(&idx_entry,sizeof(aIndex2));
    str_num=(idx_entry.num1-'0')*10+(idx_entry.num2-'0');
    if(str_num>AVI_MAX_STREAMS||str_num<0){ return 1;};/*Error*/
    if(str_num>streams) streams=str_num;
    index[str_num].type=idx_entry.type;
    index[str_num].size+=idx_entry.dwSize;
  }
  for(c=0;c<=streams;c++){
    switch(index[c].type){
    case 'w':
      a_c++;
      tmp=GetIdxNumericVar("a%d.l",a_c);
      SetIdxNumericVar("a%d.bps",a_c,8*index[c].size/(tmp?tmp:0xFFFFFFFF));
      break;
    case 'd':
      v_c++;
      tmp=GetIdxNumericVar("v%d.l",v_c);
      SetIdxNumericVar("v%d.bps",v_c,8*index[c].size/(tmp?tmp:0xFFFFFFFF));
      break;
    }
  }
  FREE(index);
  return 0;
}
#endif

#define SAFEREAD(to,size) { if(pos+size <= xsize) { memcpy(to, &xdata[pos], size); pos+= size; } else return prev;};

/* video/avi */
struct EXTRACTOR_Keywords * libextractor_riff_extract(char * filename,
						      char * xdata,
						      size_t xsize,
						      struct EXTRACTOR_Keywords * prev) {

  int size;
  int temp;
  int xtemp;
  int listtype;
  DWORD strh[16]; /*sizeof(AVISTREAMHEADER)/sizeof(DWORD)*/
  DWORD vids[12]; /*sizeof(BITMAPINFOHEADER)/sizeof(DWORD)*/
  WORD auds[13]; /*sizeof(WAVEFORMATEX)/sizeof(WORD)*/
  int stream=0;
  double fps=0; /*using for fps calculating */
  size_t pos;
  char * format;

  stream = 0;
  pos = 0;
  do {
    /*main riff chunk reading loop*/
    SAFEREAD(&temp,sizeof(int)); /*chunk type*/
    SAFEREAD(&size,sizeof(int));
    switch(temp){
    case _FOURCC_LIST:
      SAFEREAD(&listtype,sizeof(int)); 
      switch(listtype){
      case _FOURCC_hdrl:
	pos += 64; /* just a skip of MAIN  AVI HEADER */
	break;
      case _FOURCC_strl:
	SAFEREAD(strh,sizeof(strh));
	switch(strh[fccType]) {
	case _FOURCC_vids:
	  SAFEREAD(vids,sizeof(vids));
	  stream++;
	  fps=strh[dwScale]?(double)strh[dwRate]/(double)strh[dwScale]:1;
	  
	  format = malloc(256);
	  if (fps == 0) fps = 1; /* guard against div by 0 */
	  sprintf(format,
		  "video: %d.%d fps, %dm%d",	     
		  (int) fps,
		  (int)((fps-(int)fps)*AVI_FPS_NUMBERS), /* dirty hack, ne?*/
		  (int) ((strh[dwLenght]/fps) / 60), ((int)(strh[dwLenght]/fps)) % 60);
	  addKeyword(&prev,
		     format,
		     EXTRACTOR_FORMAT);
	  format = malloc(256);
	  sprintf(format,
		  "%ux%u",
		  vids[biWidth],
		  vids[biHeight]);
	  addKeyword(&prev,
		     format,
		     EXTRACTOR_SIZE);
 
	  /*
	  SetIdxNumericVar("v%d.cc",stream,vids[biCompression]);
	  */
	  pos += vids[1]+8-sizeof(vids);
	  break;
	case _FOURCC_auds:
	  SAFEREAD(auds,sizeof(auds));

	  format = malloc(256);
	  if (strh[dwScale]) 
	    xtemp = (strh[dwLenght]*strh[dwScale])/(strh[dwRate]?strh[dwRate]:0xFFFFFFFF);
	  else
	    xtemp = 0;
	  sprintf(format,
		  "audio: %u hz, %dm%d, %s",
		  auds[nSamplesPerSec]+(auds[nSamplesPerSec+1]<<16),
		  xtemp / 60, xtemp % 60,
		  auds[nChannels] > 1 ? "stereo" : "mono");
	  addKeyword(&prev,
		     format,
		     EXTRACTOR_FORMAT);
 
	  /*
	    SetIdxNumericVar("a%d.cc",stream,auds[wFormatTag]);
	  */
	  pos += auds[2]+8-sizeof(auds);
	  break;
	default: 
	  pos += 4;
	  SAFEREAD(&size,sizeof(int));
	}
	break;
      case _FOURCC_INFO:	
	prev = riffparse_INFO(xdata, xsize, prev);
	break;
      default:
	pos += size - 4; /*skip unknown LIST-TYPE */
      }
      break;
    case _FOURCC_RIFF: 
      pos += 4;
      break; 
    case _FOURCC_idx1: 
#if later
      if (!riffparse_idx1 (file, size))
	return 1; /*no index - ok, ignore index*/
#endif
      break;
    default: 
      pos += size;
    };
  } while(pos < xsize);

  return prev;
}




#if HAVE_MAIN
int main (int argc, char **argv) {
  int file;
  char * buffer;
  struct stat fstatbuf;
  size_t size;

  if (argc != 2) {
    fprintf(stderr, 
	    "Call with filename as argument\n");
    return -1;
  }  
  file = OPEN(argv[1],O_RDONLY);
  if (-1 == file) 
    return -1;
  if (-1 == fstat(file, &fstatbuf)) {
    close(file);
    return -1;
  }
  size = fstatbuf.st_size;
#ifndef MINGW
  buffer = mmap(NULL, size, PROT_READ, MAP_SHARED, file, 0);
#else
    buffer = malloc(size + 1);
    buffer[size + 1] = 0;
    read(file, buffer, size);
#endif
  close(file);

  EXTRACTOR_printKeywords(stdout,
			  libextractor_riff_extract(argv[1], 
						    buffer,
						    size,
						    NULL));
#ifndef MINGW
  munmap(buffer,size);
#else
  free(buffer);
#endif
  return 0;
}
#endif
